__all__ = ["APIError", "Connection", "HubstorageClient",
           "ScrapinghubClient", "ScrapinghubAPIError",
           "DuplicateJobError", "BadRequest", "NotFound",
           "Unauthorized", "ValueTooLarge", "ServerError"]

import pkgutil
__version__ = pkgutil.get_data(__package__, 'VERSION')
__version__ = str(__version__.decode('ascii').strip())
del pkgutil


from .legacy import *
from .hubstorage import HubstorageClient
from .client import ScrapinghubClient
from .client.exceptions import (
    ScrapinghubAPIError,
    DuplicateJobError,
    BadRequest,
    NotFound,
    Unauthorized,
    ValueTooLarge,
    ServerError,
)



from __future__ import absolute_import

from requests.compat import urljoin

from .exceptions import NotFound, _wrap_http_errors
from .jobs import Jobs
from .utils import get_tags_for_update


class Spiders(object):
    """Class to work with a collection of project spiders.

    Not a public constructor: use :class:`~scrapinghub.client.projects.Project`
    instance to get a :class:`Spiders` instance.
    See :attr:`~scrapinghub.client.projects.Project.spiders` attribute.

    :ivar project_id: string project id.

    Usage::

        >>> project.spiders
        <scrapinghub.client.spiders.Spiders at 0x1049ca630>
    """

    def __init__(self, client, project_id):
        self.project_id = project_id
        self._client = client

    def get(self, spider, **params):
        """Get a spider object for a given spider name.

        The method gets/sets spider id (and checks if spider exists).

        :param spider: a string spider name.
        :return: a spider object.
        :rtype: :class:`scrapinghub.client.spiders.Spider`

        Usage::

            >>> project.spiders.get('spider2')
            <scrapinghub.client.spiders.Spider at 0x106ee3748>
            >>> project.spiders.get('non-existing')
            NotFound: Spider non-existing doesn't exist.
        """
        project = self._client._hsclient.get_project(self.project_id)
        spider_id = project.ids.spider(spider, **params)
        if spider_id is None:
            raise NotFound("Spider {} doesn't exist.".format(spider))
        return Spider(self._client, self.project_id, spider_id, spider)

    def list(self):
        """Get a list of spiders for a project.

        :return: a list of dictionaries with spiders metadata.
        :rtype: :class:`list[dict]`

        Usage::

            >>> project.spiders.list()
            [{'id': 'spider1', 'tags': [], 'type': 'manual', 'version': '123'},
             {'id': 'spider2', 'tags': [], 'type': 'manual', 'version': '123'}]
        """
        project = self._client._connection[self.project_id]
        return project.spiders()

    def iter(self):
        """Iterate through a list of spiders for a project.

        :return: an iterator over spiders list where each spider is represented
            as a dict containing its metadata.
        :rtype: :class:`collection.Iterable[dict]`

        Provided for the sake of API consistency.
        """
        return iter(self.list())


class Spider(object):
    """Class representing a Spider object.

    Not a public constructor: use :class:`Spiders` instance to get
    a :class:`Spider` instance. See :meth:`Spiders.get` method.

    :ivar project_id: a string project id.
    :ivar key: a string key in format 'project_id/spider_id'.
    :ivar name: a spider name string.
    :ivar jobs: a collection of jobs, :class:`~scrapinghub.client.jobs.Jobs` object.

    Usage::

        >>> spider = project.spiders.get('spider1')
        >>> spider.key
        '123/1'
        >>> spider.name
        'spider1'
    """

    def __init__(self, client, project_id, spider_id, spider):
        self.project_id = project_id
        self.key = '{}/{}'.format(str(project_id), str(spider_id))
        self._id = str(spider_id)
        self.name = spider
        self.jobs = Jobs(client, project_id, self)
        self._client = client

    @_wrap_http_errors
    def update_tags(self, add=None, remove=None):
        """Update tags for the spider.

        :param add: (optional) a list of string tags to add.
        :param remove: (optional) a list of string tags to remove.
        """
        params = get_tags_for_update(add=add, remove=remove)
        path = 'v2/projects/{}/spiders/{}/tags'.format(self.project_id,
                                                       self._id)
        url = urljoin(self._client._connection.url, path)
        response = self._client._connection._session.patch(url, json=params)
        response.raise_for_status()

    @_wrap_http_errors
    def list_tags(self):
        """List spider tags.

        :return: a list of spider tags.
        :rtype: :class:`list[str]`
        """
        path = 'v2/projects/{}/spiders/{}'.format(self.project_id, self._id)
        url = urljoin(self._client._connection.url, path)
        response = self._client._connection._session.get(url)
        response.raise_for_status()
        return response.json().get('tags', [])



from __future__ import absolute_import

from .proxy import _ItemsResourceProxy


class Samples(_ItemsResourceProxy):
    """Representation of collection of job samples.

    Not a public constructor: use :class:`~scrapinghub.client.jobs.Job` instance
    to get a :class:`Samples` instance.
    See :attr:`~scrapinghub.client.jobs.Job.samples` attribute.

    Please note that :meth:`list` method can use a lot of memory and for
    a large amount of logs it's recommended to iterate through it via
    :meth:`iter` method (all params and available filters are same for
    both methods).

    Usage:

    - retrieve all samples from a job::

        >>> job.samples.iter()
        <generator object mpdecode at 0x10f5f3aa0>

    - retrieve samples with timestamp greater or equal to given timestamp::

        >>> job.samples.list(startts=1484570043851)
        [[1484570043851, 554, 576, 1777, 821, 0],
         [1484570046673, 561, 583, 1782, 821, 0]]
    """



import six
from json import dumps, loads
from datetime import datetime

EPOCH = datetime.utcfromtimestamp(0)
ADAYINSECONDS = 24 * 3600


try:
    from msgpack import Unpacker
    MSGPACK_AVAILABLE = True
except ImportError:
    MSGPACK_AVAILABLE = False


def jlencode(iterable):
    if isinstance(iterable, (dict, six.string_types)):
        iterable = [iterable]
    return u'\n'.join(jsonencode(o) for o in iterable)


def jldecode(lineiterable):
    for line in lineiterable:
        yield loads(line)


def mpdecode(iterable):
    unpacker = Unpacker()
    for chunk in iterable:
        unpacker.feed(chunk)
        # Each chunk can have none or many objects,
        # so here we dispatch any object ready
        for obj in unpacker:
            yield obj


def jsonencode(o):
    return dumps(o, default=jsondefault)


def jsondefault(o):
    if isinstance(o, datetime):
        # convert TZ-aware datetime object to POSIX timestamp
        if o.tzinfo:
            o = o.replace(tzinfo=None) - o.utcoffset()
        delta = o - EPOCH
        u = delta.microseconds
        s = delta.seconds
        d = delta.days
        return (u + (s + d * ADAYINSECONDS) * 1e6) // 1000
    else:
        return six.text_type(o)



from __future__ import absolute_import

import sys

from .proxy import _ItemsResourceProxy, _DownloadableProxyMixin


class Items(_DownloadableProxyMixin, _ItemsResourceProxy):
    """Representation of collection of job items.

    Not a public constructor: use :class:`~scrapinghub.client.jobs.Job`
    instance to get a :class:`Items` instance. See
    :attr:`~scrapinghub.client.jobs.Job.items` attribute.

    Please note that :meth:`list` method can use a lot of memory and for
    a large number of items it's recommended to iterate through them via
    :meth:`iter` method (all params and available filters are same for
    both methods).

    Usage:

    - retrieve all scraped items from a job::

        >>> job.items.iter()
        <generator object mpdecode at 0x10f5f3aa0>

    - iterate through first 100 items and print them::

        >>> for item in job.items.iter(count=100):
        ...     print(item)

    - retrieve items with timestamp greater or equal to given timestamp
      (item here is an arbitrary dictionary depending on your code)::

        >>> job.items.list(startts=1447221694537)
        [{
            'name': ['Some custom item'],
            'url': 'http://some-url/item.html',
            'size': 100000,
        }]

    - retrieve items via a generator of lists. This is most useful in cases
      where the job has a huge amount of items and it needs to be broken down
      into chunks when consumed. This example shows a job with 3 items::

        >>> gen = job.items.list_iter(chunksize=2)
        >>> next(gen)
        [{'name': 'Item #1'}, {'name': 'Item #2'}]
        >>> next(gen)
        [{'name': 'Item #3'}]
        >>> next(gen)
        Traceback (most recent call last):
          File "<stdin>", line 1, in <module>
        StopIteration

    - retrieving via meth::`list_iter` also supports the `start` and `count`.
      params. This is useful when you want to only retrieve a subset of items in
      a job. The example below belongs to a job with 10 items::

        >>> gen = job.items.list_iter(chunksize=2, start=5, count=3)
        >>> next(gen)
        [{'name': 'Item #5'}, {'name': 'Item #6'}]
        >>> next(gen)
        [{'name': 'Item #7'}]
        >>> next(gen)
        Traceback (most recent call last):
          File "<stdin>", line 1, in <module>
        StopIteration

    - retrieve 1 item with multiple filters::

        >>> filters = [("size", ">", [30000]), ("size", "<", [40000])]
        >>> job.items.list(count=1, filter=filters)
        [{
            'name': ['Some other item'],
            'url': 'http://some-url/other-item.html',
            'size': 35000,
        }]
    """

    def _modify_iter_params(self, params):
        """Modify iter filter to convert offset to start parameter.

        :return: a dict with updated set of params.
        :rtype: :class:`dict`
        """
        params = super(Items, self)._modify_iter_params(params)
        offset = params.pop('offset', None)
        if offset:
            params['start'] = '{}/{}'.format(self.key, offset)
        return params

    def list_iter(self, chunksize=1000, *args, **kwargs):
        """An alternative interface for reading items by returning them
        as a generator which yields lists of items sized as `chunksize`.

        This is a convenient method for cases when processing a large amount of
        items from a job isn't ideal in one go due to the large memory needed.
        Instead, this allows you to process it chunk by chunk.

        You can improve I/O overheads by increasing the chunk value but that
        would also increase the memory consumption.

        :param chunksize: size of list to be returned per iteration
        :param start: offset to specify the start of the item iteration
        :param count: overall number of items to be returned, which is broken
            down by `chunksize`.

        :return: an iterator over items, yielding lists of items.
        :rtype: :class:`collections.Iterable`
        """

        start = kwargs.pop("start", 0)
        count = kwargs.pop("count", sys.maxsize)
        processed = 0

        while True:
            next_key = self.key + "/" + str(start)
            if processed + chunksize > count:
                chunksize = count - processed
            items = [
                item for item in self.iter(
                    count=chunksize, start=next_key, *args, **kwargs)
            ]
            yield items
            processed += len(items)
            start += len(items)
            if processed >= count:
                break
            if len(items) < chunksize:
                break



import sys
from os.path import dirname, join

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


with open(join(dirname(__file__), 'scrapinghub/VERSION'), 'rb') as f:
    version = f.read().decode('ascii').strip()

is_pypy = '__pypy__' in sys.builtin_module_names
mpack_required = ['msgpack>=1.0.0']
if is_pypy:
    mpack_required.append('msgpack-pypy>=0.0.2')

setup(
    name='scrapinghub',
    version=version,
    license='BSD',
    description='Client interface for Scrapinghub API',
    author='Scrapinghub',
    author_email='info@scrapinghub.com',
    url='http://github.com/scrapinghub/python-scrapinghub',
    platforms=['Any'],
    packages=['scrapinghub', 'scrapinghub.client', 'scrapinghub.hubstorage'],
    package_data={'scrapinghub': ['VERSION']},
    install_requires=['requests>=1.0', 'retrying>=1.3.3', 'six>=1.10.0'],
    extras_require={'msgpack': mpack_required},
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Internet :: WWW/HTTP',
    ],
)
